package cc.fascinated.backend.model.score;

import cc.fascinated.backend.common.DateUtils;
import cc.fascinated.backend.leaderboard.impl.ScoreSaberLeaderboard;
import cc.fascinated.backend.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.backend.model.token.ScoreSaberPlayerScoreToken;
import cc.fascinated.backend.model.token.ScoreSaberScoreToken;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.Setter;
import org.springframework.data.annotation.Id;
import org.springframework.data.mongodb.core.mapping.Document;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * A score for an account.
 */
@Document
@Getter @Setter
public class Score {
    /**
     * The id for this score.
     */
    @Id
    private String id;

    /**
     * The rank of this score.
     */
    private int rank;

    /**
     * The base score for this score.
     */
    private int baseScore;

    /**
     * The modified score for this score.
     */
    private int modifiedScore;

    /**
     * The PP for this score.
     */
    private Double pp;

    /**
     * The weight for this score.
     */
    private int weight;

    /**
     * The modifiers for this score.
     */
    private List<String> modifiers;

    /**
     * The multiplier for this score.
     */
    private int multiplier;

    /**
     * How many bad cuts this score has.
     */
    private int badCuts;

    /**
     * How many misses this score has.
     */
    private int missedNotes;

    /**
     * The maximum combo for this score.
     */
    private int maxCombo;

    /**
     * Whether this score was a full combo.
     */
    private boolean fullCombo;

    /**
     * The HMD that was used to set this score.
     */
    private int hmd;

    /**
     * The time set for this score.
     */
    private Date timeSet;

    /**
     * Whether this score has a replay.
     */
    private boolean hasReplay;

    /**
     * The full HMD name that was used to set this score.
     */
    private String deviceHmd;

    /**
     * The controller that was used on the left hand.
     */
    private String deviceControllerLeft;

    /**
     * The controller that was used on the right hand.
     */
    private String deviceControllerRight;

    /**
     * The previous scores for this score.
     */
    private List<Score> previousScores;

    /**
     * The account who set this score.
     */
    private String accountId;

    /**
     * The leaderboard id for this score.
     */
    private String leaderboardId;

    /**
     * The curve version for this score.
     */
    private Integer curveVersion;

    /**
     * The difficulty this score was set on.
     */
    private Difficulty difficulty;

    public Score(String id, int rank, int baseScore, int modifiedScore, Double pp, int weight, List<String> modifiers,
                 int multiplier, int badCuts, int missedNotes, int maxCombo, boolean fullCombo, int hmd, Date timeSet,
                 boolean hasReplay, String deviceHmd, String deviceControllerLeft, String deviceControllerRight,
                 List<Score> previousScores, String accountId, String leaderboardId, Integer curveVersion, Difficulty difficulty) {
        this.id = id;
        this.rank = rank;
        this.baseScore = baseScore;
        this.modifiedScore = modifiedScore;
        this.pp = pp;
        this.weight = weight;
        this.modifiers = modifiers;
        this.multiplier = multiplier;
        this.badCuts = badCuts;
        this.missedNotes = missedNotes;
        this.maxCombo = maxCombo;
        this.fullCombo = fullCombo;
        this.hmd = hmd;
        this.timeSet = timeSet;
        this.hasReplay = hasReplay;
        this.deviceHmd = deviceHmd;
        this.deviceControllerLeft = deviceControllerLeft;
        this.deviceControllerRight = deviceControllerRight;
        this.previousScores = previousScores;
        this.accountId = accountId;
        this.leaderboardId = leaderboardId;
        this.curveVersion = curveVersion == null ? ScoreSaberLeaderboard.INSTANCE.getCurveVersion() : curveVersion;
        this.difficulty = difficulty;
    }

    /**
     * Gets a score from the given token.
     *
     * @param token The token.
     * @return The score.
     */
    public static Score fromToken(String playerId, ScoreSaberPlayerScoreToken token) {
        ScoreSaberScoreToken score = token.getScore();
        ScoreSaberLeaderboardToken leaderboard = token.getLeaderboard();
        List<String> modifiers = new ArrayList<>(List.of(score.getModifiers().split(",")));
        // If the token's modifiers aren't a list, add the only modifier.
        if (modifiers.isEmpty() && !score.getModifiers().isEmpty()) {
            modifiers.add(score.getModifiers());
        }

        // Return the score.
        return new Score(
                score.getId(),
                score.getRank(),
                score.getBaseScore(),
                score.getModifiedScore(),
                score.getPp(),
                score.getWeight(),
                modifiers,
                score.getMultiplier(),
                score.getBadCuts(),
                score.getMissedNotes(),
                score.getMaxCombo(),
                score.isFullCombo(),
                score.getHmd(),
                DateUtils.getDateFromString(score.getTimeSet()),
                score.isHasReplay(),
                score.getDeviceHmd(),
                score.getDeviceControllerLeft(),
                score.getDeviceControllerRight(),
                new ArrayList<>(),
                playerId,
                leaderboard.getId(),
                ScoreSaberLeaderboard.INSTANCE.getCurveVersion(), // Get the curve version from the leaderboard.
                Difficulty.fromId(leaderboard.getDifficulty().getDifficulty())
        );
    }

    @AllArgsConstructor @Getter
    public enum Difficulty {
        EASY(1),
        NORMAL(3),
        HARD(5),
        EXPERT(7),
        EXPERT_PLUS(9);

        /**
         * The ScoreSaber difficulty id.
         */
        private final int id;

        /**
         * Gets the difficulty from the given id.
         *
         * @param id The id.
         * @return The difficulty.
         */
        public static Difficulty fromId(int id) {
            for (Difficulty difficulty : values()) {
                if (difficulty.getId() == id) {
                    return difficulty;
                }
            }
            return null;
        }
    }
}

